#include "stdafx.h"
#include "clf_camera.h"

void cCamera::moveby(float fwdx,float fwdy,float fwdz)
{
	vLocation.x += fwdx;
	vLocation.y += fwdy;
	vLocation.z += fwdz;
}

clf_matrix4x4 cCamera::rotation_matrix()
{
	clf_matrix4x4 mMatrix;
	clf_vector3D vAxisX;
	clf_vector3D zFlipped;

	zFlipped.x = -vForward.x;
	zFlipped.y = -vForward.y;
	zFlipped.z = -vForward.z;

	// Derive X vector
	vAxisX = crossProduct(zFlipped, vUp);

	// Populate matrix, note this is just the rotation and is transposed
	mMatrix[0] = vAxisX.x;
	mMatrix[4] = vAxisX.y;
	mMatrix[8] = vAxisX.z;
	mMatrix[12] = 0.0f;

	mMatrix[1] = vUp.x;
	mMatrix[5] = vUp.y;
	mMatrix[9] = vUp.z;
	mMatrix[13] = 0.0f;

	mMatrix[2] = zFlipped.x;
	mMatrix[6] = zFlipped.y;
	mMatrix[10] = zFlipped.z;
	mMatrix[14] = 0.0f;

	mMatrix[3] = 0.0f;
	mMatrix[7] = 0.0f;
	mMatrix[11] = 0.0f;
	mMatrix[15] = 1.0f;

	return mMatrix;
}

void cCamera::apply()
{
	clf_matrix4x4 mMatrix;
	clf_vector3D vAxisX;
	clf_vector3D zFlipped;

	zFlipped.x = -vForward.x;
	zFlipped.y = -vForward.y;
	zFlipped.z = -vForward.z;

	// Derive X vector
	vAxisX = crossProduct(zFlipped, vUp);

	// Populate matrix, note this is just the rotation and is transposed
	mMatrix[0] = vAxisX.x;
	mMatrix[4] = vAxisX.y;
	mMatrix[8] = vAxisX.z;
	mMatrix[12] = 0.0f;

	mMatrix[1] = vUp.x;
	mMatrix[5] = vUp.y;
	mMatrix[9] = vUp.z;
	mMatrix[13] = 0.0f;

	mMatrix[2] = zFlipped.x;
	mMatrix[6] = zFlipped.y;
	mMatrix[10] = zFlipped.z;
	mMatrix[14] = 0.0f;

	mMatrix[3] = 0.0f;
	mMatrix[7] = 0.0f;
	mMatrix[11] = 0.0f;
	mMatrix[15] = 1.0f;

	// Do the rotation first
	glMultMatrixf(mMatrix);

	// Now, translate backwards
	glTranslatef(-vLocation.x, -vLocation.y, -vLocation.z);
}

void cCamera::apply_mirror()
{
	clf_matrix4x4 mMatrix;
	clf_vector3D vAxisX;
	clf_vector3D zFlipped;

	zFlipped.x = vForward.x;
	zFlipped.y = vForward.y;
	zFlipped.z = vForward.z;

	// Derive X vector
	vAxisX = -crossProduct(zFlipped, vUp);

	// Populate matrix, note this is just the rotation and is transposed
	mMatrix[0] = vAxisX.x;
	mMatrix[4] = vAxisX.y;
	mMatrix[8] = vAxisX.z;
	mMatrix[12] = 0.0f;

	mMatrix[1] = vUp.x;
	mMatrix[5] = vUp.y;
	mMatrix[9] = vUp.z;
	mMatrix[13] = 0.0f;

	mMatrix[2]  = zFlipped.x;
	mMatrix[6]  = zFlipped.y;
	mMatrix[10] = zFlipped.z;
	mMatrix[14] = .0f;

	mMatrix[3] = .0f;
	mMatrix[7] = .0f;
	mMatrix[11] =.0f;
	mMatrix[15] = 1.0f;

	// Do the rotation first
	glMultMatrixf(mMatrix);

	// Now, translate backwards
	glTranslatef(-vLocation.x, -vLocation.y, -vLocation.z);
}

void cCamera::rotate(float x, float y, float z)
{
	// x - forward
	// y - up
	// z - right
	rotate_locZ(x);
	rotate_locX(y);
	rotate_locY(z);
}

void cCamera::rotate_vec(clf_vector3D& v, float x, float y, float z)
{
	// rotate in local coordinates
	rotate_point(x, y, z, v);

	float xp,yp,zp;
	clf_vector3D vAxisY;
	clf_vector3D zFlipped;

	// get Y axis
	zFlipped.x = -vForward.x;
	zFlipped.y = -vForward.y;
	zFlipped.z = -vForward.z;
	vAxisY = crossProduct(zFlipped, vUp);

	// convert to global coordinate system
	xp = v.x*vForward.x + vAxisY.x*v.y + vUp.x*v.z;
	yp = v.x*vForward.y + vAxisY.y*v.y + vUp.y*v.z;
	zp = v.x*vForward.z + vAxisY.z*v.y + vUp.z*v.z;

	v = clf_vector3D(xp, yp, zp);
}

void cCamera::rotate_locX(float fAngle)
{
	clf_matrix4x4 mRotation;
	clf_vector3D vCross;

	vCross = crossProduct(vUp, vForward);
	mRotation.RotationMatrix(fAngle, vCross.x, vCross.y, vCross.z);

	clf_vector3D vNewVect;
	// Inline 3x3 matrix multiply for rotation only
	vNewVect.x = mRotation[0] * vForward.x + mRotation[4] * vForward.y + mRotation[8] *  vForward.z;	
	vNewVect.y = mRotation[1] * vForward.x + mRotation[5] * vForward.y + mRotation[9] *  vForward.z;	
	vNewVect.z = mRotation[2] * vForward.x + mRotation[6] * vForward.y + mRotation[10] * vForward.z;	
	memcpy(vForward, vNewVect, sizeof(float)*3);

	// Update pointing up vector
	vNewVect.x = mRotation[0] * vUp.x + mRotation[4] * vUp.y + mRotation[8] *  vUp.z;	
	vNewVect.y = mRotation[1] * vUp.x + mRotation[5] * vUp.y + mRotation[9] *  vUp.z;	
	vNewVect.z = mRotation[2] * vUp.x + mRotation[6] * vUp.y + mRotation[10] * vUp.z;	
	memcpy(vUp, vNewVect, sizeof(float) * 3);
}

void cCamera::rotate_locY(float angle)
{
	clf_matrix4x4 mRotation;

	mRotation.RotationMatrix(angle, 0,1,0);
	mRotation.RotationMatrix(angle, vUp.x, vUp.y, vUp.z);

	vForward.rotate(mRotation);
}

void cCamera::rotate_locZ(float fAngle)
{
	clf_matrix4x4 mRotation;

	// Only the up vector needs to be rotated
	mRotation.RotationMatrix(fAngle, vForward.x, vForward.y, vForward.z);

	clf_vector3D vNewVect;
	vNewVect.x = mRotation[0] * vUp.x + mRotation[4] * vUp.y + mRotation[8] *  vUp.z;	
	vNewVect.y = mRotation[1] * vUp.x + mRotation[5] * vUp.y + mRotation[9] *  vUp.z;	
	vNewVect.z = mRotation[2] * vUp.x + mRotation[6] * vUp.y + mRotation[10] * vUp.z;	
	memcpy(vUp, vNewVect, sizeof(float) * 3);
}

void cCamera::setup(clf_vector3D& pos, clf_vector3D& lookat, clf_vector3D& up)
{
	vLocation = pos;
	vForward = lookat-pos;
	vUp = up;
	vForward.normalize();

	clf_vector3D right = crossProduct(vForward, vUp);
	vUp = crossProduct(right, vForward);
	vUp.normalize();
}
